/*
 * Copyright (C) 2012-2013, Eric Wong <normalperson@yhbt.net>
 * License: GPLv3 or later (see COPYING for details)
 */
#include "cmogstored.h"
#include "cfg.h"
#include "nostd/setproctitle.h"
#define THIS "cmogstored"
#define DESC "alternative mogstored implementation for MogileFS"
static char summary[] = THIS " -- "DESC;
const char *argp_program_bug_address = PACKAGE_BUGREPORT;
const char *argp_program_version = THIS" "PACKAGE_VERSION;
static sig_atomic_t sigchld_nr;
static sig_atomic_t do_exit;
static size_t nthr;
static bool have_mgmt;
static pid_t master_pid;
static unsigned long worker_processes;
static Hash_table *workers;
struct worker {
	pid_t pid;
	unsigned long id;
};

#define CFG_KEY(f) -((int)offsetof(struct mog_cfg,f) + 1)
static struct argp_option options[] = {
	{ .name = "daemonize", .key = 'd',
	  .doc = "Daemonize" },
	{ .name = "config", .key = CFG_KEY(configfile),
	  .arg = "<file>",
          .doc = "Set config file (no default, unlike mogstored)" },
	{ .name = "httplisten", .key = CFG_KEY(httplisten),
	  .arg = "<ip:port>",
	  .doc = "IP/Port HTTP server listens on" },
	/*
	 * NOT adding httpgetlisten here to avoid (further) breaking
	 * compatibility with Perl mogstored.  Most of these command-line
	 * options suck anyways.
	 */
	{ .name = "mgmtlisten", .key = CFG_KEY(mgmtlisten),
	  .arg = "<ip:port>",
	  .doc = "IP/Port management/sidechannel listens on" },
	{ .name = "docroot", .key = CFG_KEY(docroot),
	  .arg = "<path>",
	  .doc = "Docroot above device mount points.  "
	         "Defaults to "MOG_DEFAULT_DOCROOT
	},
	{ .name = "maxconns", .key = CFG_KEY(maxconns),
	  .arg = "<number>",
	  .doc = "Number of simultaneous clients to serve. "
	         "Default " MOG_STR(MOG_DEFAULT_MAXCONNS) },
	{ .name = "pidfile", .key = CFG_KEY(pidfile),
	  .arg = "<path>",
	  .doc = "path to PID file" },
	{ .name = "server", .key = CFG_KEY(server), .flags = OPTION_HIDDEN,
	  .arg = "(perlbal|none)"
	},
	{
	  /* hidden for now, don't break compat with Perl mogstored */
	  .name = "multi", .key = -'M', .flags = OPTION_HIDDEN
	},
	{
	  /* hidden for now, don't break compat with Perl mogstored */
	  .name = "worker-processes", .key = -'W', .flags = OPTION_HIDDEN,
	  .arg = "<number>",
	},
	/* we don't load a default config file like Perl mogstored at all */
	{ .name = "skipconfig", .key = 's', .flags = OPTION_HIDDEN },
	{ NULL }
};

static void new_cfg_or_die(const char *config)
{
	struct mog_cfg *cfg = mog_cfg_new(config);

	if (!cfg) die("invalid (or duplicate) config=%s", config);
	if (mog_cfg_load(cfg) == 0) return;

	die("failed to load config=%s: %s",
	    config, errno ? strerror(errno) : "parser error");
}

static void check_strtoul(unsigned long *dst, const char *s, const char *key)
{
	char *end;

	errno = 0;
	*dst = strtoul(s, &end, 10);
	if (errno)
		die_errno("failed to parse --%s=%s", key, s);
	if (*end)
		die("failed to parse --%s=%s (invalid character: %c)",
		    key, s, *end);
}

static void addr_or_die(struct mog_addrinfo **dst, const char *key, char *s)
{
	*dst = mog_listen_parse(s);
	if (!*dst)
		die("failed to parse %s=%s", key, s);
}

static error_t parse_opt(int key, char *arg, struct argp_state *state)
{
	struct mog_cfg *cfg = state->input;
	int rv = 0;

	switch (key) {
	case 'd': cfg->daemonize = 1; break;
	case 's': /* no-op, we don't load the default config file */; break;
	case CFG_KEY(docroot): cfg->docroot = xstrdup(arg); break;
	case CFG_KEY(pidfile): cfg->pidfile = xstrdup(arg); break;
	case CFG_KEY(configfile): new_cfg_or_die(arg); break;
	case CFG_KEY(maxconns):
		check_strtoul(&cfg->maxconns, arg, "maxconns");
		break;
	case CFG_KEY(httplisten):
		addr_or_die(&cfg->httplisten, "httplisten", arg);
		break;
	case CFG_KEY(mgmtlisten):
		addr_or_die(&cfg->mgmtlisten, "mgmtlisten", arg);
		break;
	case CFG_KEY(server):
		cfg->server = xstrdup(arg);
		mog_cfg_check_server(cfg);
		break;
	case -'M': mog_cfg_multi = true; break;
	case -'W':
		check_strtoul(&worker_processes, arg, "worker-processes");
		break;
	case ARGP_KEY_ARG:
		argp_usage(state);
	case ARGP_KEY_END:
		break;
	default:
		rv = ARGP_ERR_UNKNOWN;
	}

	return rv;
}

static int open_pidfile(void)
{
	pid_t cur_pid = -1;
	int pid_fd = mog_pidfile_open(mog_cli.pidfile, &cur_pid);

	if (pid_fd >= 0)
		return pid_fd;
	if (errno == EAGAIN)
		die("already running on PID: %d", (int)cur_pid);
	else
		die_errno("mog_pidfile_open failed");
	return -1;
}

static void dup2_null(int oldfd, int newfd, const char *errdest)
{
	int rc;

	do {
		rc = dup2(oldfd, newfd);
	} while (rc < 0 && (errno == EINTR || errno == EBUSY));

	if (rc < 0)
		die_errno("dup2(/dev/null,%s) failed", errdest);
}

static void daemonize(void)
{
	int ready_pipe[2];
	pid_t pid;
	ssize_t r;
	int fd;

	if (pipe(ready_pipe) < 0)
		die_errno("pipe() failed");
	pid = fork();
	if (pid < 0)
		die_errno("fork() failed");
	if (pid > 0) { /* parent */
		mog_close(ready_pipe[1]);
		do {
			r = read(ready_pipe[0], &pid, sizeof(pid_t));
		} while (r < 0 && errno == EINTR);

		PRESERVE_ERRNO( mog_close(ready_pipe[0]) );
		if (r == sizeof(pid_t))
			exit(EXIT_SUCCESS);
		if (r < 0)
			die_errno("ready_pipe read error");
		die("daemonization error, check syslog");
	}

	/* child */
	mog_close(ready_pipe[0]);
	if (setsid() < 0)
		die_errno("setsid() failed");

	pid = fork();
	if (pid < 0)
		die_errno("fork() failed");
	if (pid > 0) /* intermediate parent */
		exit(EXIT_SUCCESS);

	if (chdir("/") < 0)
		die_errno("chdir(/) failed");
	fd = open("/dev/null", O_RDWR);
	if (fd < 0)
		die_errno("open(/dev/null) failed");

	dup2_null(fd, STDIN_FILENO, "stdin");
	dup2_null(fd, STDOUT_FILENO, "stdout");
	dup2_null(fd, STDERR_FILENO, "stderr");
	mog_close(fd);

	do {
		r = write(ready_pipe[1], &pid, sizeof(pid_t));
	} while (r < 0 && errno == EINTR);
	if (r < 0)
		syslog(LOG_CRIT, "ready_pipe write error: %m");
	else
		assert(r == sizeof(pid_t) && "impossible short write");
	mog_close(ready_pipe[1]);
}

#ifndef LOG_PERROR
# define LOG_PERROR 0
#endif

MOG_NOINLINE static void setup(int argc, char *argv[])
{
	int pid_fd = -1;
	static struct argp argp = { options, parse_opt, NULL, summary };
	int mask = 0;

	mog_mnt_refresh();
	argp_parse(&argp, argc, argv, 0, NULL, &mog_cli);
	mog_cfg_validate_or_die(&mog_cli);
	mog_cfg_svc_start_or_die(&mog_cli);

	if (mog_cli.pidfile) pid_fd = open_pidfile();

	/* TODO: make logging configurable */
	{
		int option = LOG_PID;

		if (!mog_cli.daemonize)
			option |= LOG_PERROR;

		openlog(THIS, option, LOG_DAEMON);
		mask |= LOG_MASK(LOG_EMERG);
		mask |= LOG_MASK(LOG_ALERT);
		mask |= LOG_MASK(LOG_CRIT);
		mask |= LOG_MASK(LOG_ERR);
		mask |= LOG_MASK(LOG_WARNING);
		mask |= LOG_MASK(LOG_NOTICE);
		mask |= LOG_MASK(LOG_INFO);
		/* mask |= LOG_MASK(LOG_DEBUG); */
		setlogmask(mask);
	}

	if (mog_cli.daemonize)
		daemonize();

	if (pid_fd >= 0 && mog_pidfile_commit(pid_fd) < 0)
		syslog(LOG_ERR,
		       "failed to write pidfile(%s): %m. continuing...",
		       mog_cli.pidfile);

	master_pid = getpid();

	/* 10 - 100 threads based on number of devices, same as mogstored */
	nthr = mog_mkusage_all() * 10;
	nthr = MAX(10, nthr);
	nthr = MIN(100, nthr);
}

/* Hash iterator function */
static bool svc_start_each(void *svcptr, void *qptr)
{
	struct mog_svc *svc = svcptr;
	struct mog_queue *q = qptr;
	struct mog_accept *ac;
	size_t athr = (size_t)num_processors(NPROC_CURRENT);

	/*
	 * try to distribute accept() callers between workers more evenly
	 * with wake-one accept() behavior by trimming down on acceptors
	 */
	if (worker_processes) {
		athr /= worker_processes;
		if (athr == 0)
			athr = 1;
	}

	svc->queue = q;

	if (svc->mgmt_fd >= 0) {
		have_mgmt = true;
		ac = mog_accept_init(svc->mgmt_fd, svc, mog_mgmt_post_accept);

		/*
		 * mgmt port is rarely used and always persistent, so it
		 * does not need multiple threads for blocking accept()
		 */
		mog_thrpool_start(&ac->thrpool, 1, mog_accept_loop, ac);
	}

	if (svc->http_fd >= 0) {
		ac = mog_accept_init(svc->http_fd, svc, mog_http_post_accept);
		mog_thrpool_start(&ac->thrpool, athr, mog_accept_loop, ac);
	}

	if (svc->httpget_fd >= 0) {
		ac = mog_accept_init(svc->httpget_fd, svc,
		                     mog_httpget_post_accept);
		mog_thrpool_start(&ac->thrpool, athr, mog_accept_loop, ac);
	}

	return true;
}

static void acceptor_quit(int fd)
{
	if (fd >= 0) {
		struct mog_fd *mfd = mog_fd_get(fd);
		struct mog_accept *ac = &mfd->as.accept;

		mog_thrpool_quit(&ac->thrpool, NULL);
		mog_fd_put(mfd);
	}
}

static bool svc_quit_accept_i(void *svcptr, void *ignored)
{
	struct mog_svc *svc = svcptr;

	acceptor_quit(svc->mgmt_fd);
	acceptor_quit(svc->http_fd);
	acceptor_quit(svc->httpget_fd);

	return true;
}

static bool svc_queue_set(void *svcptr, void *queue)
{
	struct mog_svc *svc = svcptr;

	svc->queue = queue;

	return true;
}

_Noreturn static void cmogstored_exit(void)
{
	/* call atexit() handlers and make valgrind happy */
	setproctitle("cmogstored, shutting down");
	mog_svc_each(svc_quit_accept_i, NULL);
	mog_svc_dev_shutdown();
	mog_queue_stop(mog_notify_queue);
	mog_svc_dev_shutdown();
	mog_svc_each(svc_queue_set, mog_notify_queue);
	mog_fdmap_requeue(mog_notify_queue);
	mog_queue_quit_loop(mog_notify_queue);
	exit(EXIT_SUCCESS);
}

static void worker_wakeup_handler(int signum)
{
	switch (signum) {
	case SIGCHLD: ++sigchld_nr; break;
	case SIGQUIT:
	case SIGTERM:
	case SIGINT:
		do_exit = signum;
	}
	mog_notify(MOG_NOTIFY_SIGNAL);
}

static void wakeup_noop(int signum)
{
	/* just something to cause EINTR */
}

static bool kill_child(void *workent, void *ignored)
{
	struct worker *worker = workent;

	if (kill(worker->pid, do_exit) == 0)
		return true;

	/* race between receiving a signal and waitpid2, ignore */
	if (errno == ESRCH)
		return true;
	syslog(LOG_ERR, "could not signal worker[%lu] pid=%d: %m",
	       worker->id, (int)worker->pid);
	return true;
}

static void master_wakeup_handler(int signum)
{
	switch (signum) {
	case SIGQUIT:
	case SIGTERM:
	case SIGINT:
		do_exit = signum;
	}
}

static void siginit(void (*wakeup_handler)(int))
{
	struct sigaction sa;

	memset(&sa, 0, sizeof(struct sigaction));
	CHECK(int, 0, sigemptyset(&sa.sa_mask) );

	sa.sa_handler = SIG_IGN;
	CHECK(int, 0, sigaction(SIGPIPE, &sa, NULL));

	sa.sa_handler = wakeup_noop;
	CHECK(int, 0, sigaction(SIGURG, &sa, NULL));

	sa.sa_handler = wakeup_handler;
	CHECK(int, 0, sigaction(SIGTERM, &sa, NULL));
	CHECK(int, 0, sigaction(SIGINT, &sa, NULL));

	/* TODO: graceful shutdown, upgrade, reload */
	CHECK(int, 0, sigaction(SIGQUIT, &sa, NULL));
	CHECK(int, 0, sigaction(SIGHUP, &sa, NULL));
	CHECK(int, 0, sigaction(SIGUSR1, &sa, NULL));
	CHECK(int, 0, sigaction(SIGUSR2, &sa, NULL));

	sa.sa_flags = SA_NOCLDSTOP;
	CHECK(int, 0, sigaction(SIGCHLD, &sa, NULL));
}

static void main_loop(const pid_t parent)
{
	mog_cancel_disable(); /* mog_idleq_wait() now relies on this */
	while (parent == 0 || parent == getppid()) {
		mog_notify_wait(have_mgmt);
		if (do_exit)
			cmogstored_exit();
		if (have_mgmt && sigchld_nr > 0) {
			mog_mnt_refresh();

			/*
			 * mog_iostat_spawn may fail due to lack of FDs,
			 * don't reset sigchld_nr counter if it fails
			 */
			if (mog_iostat_spawn())
				sigchld_nr = 0;
		}
	}

	syslog(LOG_INFO, "parent=%d abandoned us, dying", parent);
	cmogstored_exit();
}

static void run_worker(const pid_t parent)
{
	struct mog_queue *q = mog_queue_new();

	mog_notify_init();
	siginit(worker_wakeup_handler);
	mog_thrpool_start(&q->thrpool, nthr, mog_queue_loop, q);
	have_mgmt = false;
	mog_svc_each(svc_start_each, q); /* this will set have_mgmt */
	if (have_mgmt && ! mog_iostat_spawn()) {
		sigchld_nr = 1; /* enables retrying mog_iostat_spawn() */
		syslog(LOG_WARNING, "iostat(1) not available");
	}
	main_loop(parent);
}

static bool worker_cmp(const void *a, const void *b)
{
	const struct worker *wa = a;
	const struct worker *wb = b;

	return wa->pid == wb->pid;
}

static size_t worker_hash(const void *x, size_t tablesize)
{
	const struct worker *w = x;

	return w->pid % tablesize;
}

static void fork_worker(unsigned long worker_id)
{
	struct worker *worker;
	pid_t pid;
	pid_t parent = getpid(); /* not using getppid() since it's racy */

	mog_intr_disable();
	pid = fork();
	if (pid > 0) {
		worker = xmalloc(sizeof(struct worker));
		worker->pid = pid;
		worker->id = worker_id;
		if (hash_insert(workers, worker) == NULL)
			mog_oom();
		mog_intr_enable();
	} else if (pid == 0) {
		/* worker will call mog_intr_enable() later in notify loop */
		hash_free(workers);
		run_worker(parent);
		exit(EXIT_SUCCESS);
	} else {
		PRESERVE_ERRNO( mog_intr_enable() );
		syslog(LOG_ERR, "fork() failed: %m, sleeping for 10s");
		sleep(10);
	}
}

static void worker_died(int status, struct worker *tmp)
{
	struct worker *died = hash_delete(workers, tmp);

	/*
	 * we got an unexpected pid, just ignore it, it could
	 * be a reparented iostat at shutdown
	 */
	if (died == NULL)
		return;

	tmp->id = died->id;
	free(died);

	if (do_exit)
		return;

	syslog(LOG_INFO,
	       "worker[%lu] pid=%d died with status=%d, respawning",
	       tmp->id, (int)tmp->pid, status);
	fork_worker(tmp->id);
}

static void run_master(void)
{
	unsigned long id;

	workers = hash_initialize(worker_processes, NULL,
	                              worker_hash, worker_cmp, free);

	siginit(master_wakeup_handler);
	mog_intr_enable();

	for (id = 0; id < worker_processes; id++)
		fork_worker(id);

	while (! do_exit || hash_get_n_entries(workers) > 0) {
		int status;
		struct worker tmp;

		tmp.pid = waitpid(-1, &status, 0);

		if (tmp.pid > 0)
			worker_died(status, &tmp);
		if (tmp.pid < 0 && errno != EINTR)
			syslog(LOG_WARNING, "waitpid failed: %m");

		if (do_exit && do_exit != SIGCONT) {
			hash_do_for_each(workers, kill_child, NULL);
			do_exit = SIGCONT; /* only kill once */
		}
	}
	hash_free(workers);
}

int main(int argc, char *argv[], char *envp[])
{
	/* hack for older gcov + gcc, see nostd/setproctitle.h */
	spt_init(argc, argv, envp);
	set_program_name(argv[0]);

	mog_intr_disable();
	setup(argc, argv); /* this daemonizes */

	if (worker_processes == 0)
		run_worker(0);
	else
		run_master();

	return 0;
}

/* called by the "shutdown" command via mgmt */
void cmogstored_quit(void)
{
	if (worker_processes > 0) {
		if (kill(master_pid, SIGQUIT) != 0)
			syslog(LOG_ERR,
			       "SIGQUIT failed on master process (pid=%d): %m",
				master_pid);
	}
	worker_wakeup_handler(SIGQUIT);
}
